"""Tests for status module."""

import pytest

from quietex.lexer import IO, Generic, State, Text
from quietex.status import AppState


def test_create_appstate():
    """Check initial values on new AppState."""
    state = AppState()
    assert state.current_page is None
    assert state.current_file is None


# Status bar formatting


def test_format_status_empty():
    """Test formatting status bar with no current page and file."""
    state = AppState()
    assert state.format_status() == ""


def test_format_status_page_only():
    """Test formatting status bar with only a current page."""
    state = AppState(current_page=1)
    assert state.format_status() == "[1]"


def test_format_status_file_only():
    """Test formatting status bar with only a current file."""
    state = AppState(file_stack=["./test.tex"])
    assert state.format_status() == "(./test.tex)"


def test_format_status_page_and_file():
    """Test formatting status bar with a current page and file."""
    state = AppState(current_page=1, file_stack=["./test.tex"])
    assert state.format_status() == "[1] (./test.tex)"


# Updating state from tokens


@pytest.mark.parametrize(
    "token",
    (
        (Generic.Error, "error"),
        (Text, " text "),
        (State.EndPage, "]"),
        (IO.ReadAux, "{./aux.map}"),
        (IO.ReadImage, "<./image.png>"),
        (Generic.Warning, "warning"),
    ),
)
def test_next_state_no_change(token):
    """Test that most tokens don't change the state."""
    state = AppState()
    next_state = state.update([token])
    assert next_state == state


@pytest.mark.parametrize(
    "tokens",
    (
        [(State.StartPage, "[2 ")],
        [(State.StartPage, "[1"), (State.EndPage, "]"), (State.StartPage, "[2 ")],
    ),
)
def test_next_state_page(tokens):
    """Test the current page changes when StartPage tokens are seen."""
    state = AppState()
    next_state = state.update(tokens)
    assert next_state.current_page == 2
    assert next_state != state


@pytest.mark.parametrize(
    "tokens,expected_file",
    (
        ([(IO.OpenFile, "(./test.tex")], "./test.tex"),
        ([(IO.OpenFile, "(./test.tex"), (IO.OpenFile, "(./test2.tex")], "./test2.tex"),
    ),
)
def test_next_state_file(tokens, expected_file):
    """Test the current file changes when OpenFile and CloseFile tokens are seen."""
    state = AppState()
    next_state = state.update(tokens)
    assert next_state.current_file == expected_file
    assert next_state != state


@pytest.mark.parametrize(
    "tokens", ([(IO.OpenFile, "(./test.tex"), (IO.CloseFile, ")")],)
)
def test_next_state_empty_stack(tokens):
    """Test emptying the file stack."""
    state = AppState()
    next_state = state.update(tokens)
    assert next_state.current_file is None
    assert next_state == state



"""Pygments lexer for LaTeX compiler log output."""
import re
from typing import Any, List, Tuple

from pygments.lexer import RegexLexer, default, include
from pygments.token import Generic, Text

__all__ = ["LatexLogLexer", "lex", "split"]

# Tokens
IO = Generic.IO
State = Generic.State  # pylint: disable=C0103
UI = Generic.UI
__all__ += ["Generic", "IO", "UI", "State", "Text"]


class LatexLogLexer(RegexLexer):
    """Lexer for a single line (or section thereof) of LaTeX compiler log output."""

    name = "LatexLog"

    START_PAGE_RE = r"\[\d+\s?"
    OPEN_FILE_RE = r"\(\.?/[^\s(){}]+"

    def __init__(self, ensurenl=False, **options):
        super().__init__(ensurenl=ensurenl, **options)

    def text_and_close_files(self, match):
        """Callback to lex text followed by close-files."""
        tokens = []
        text = match.group()
        if text.count("(") != text.count(")"):
            # Split trailing spaces into a separate token
            # TODO: use bygroups for this instead?
            text_stripped = text.rstrip(" ")
            if text_stripped != text:
                start = len(text_stripped)
                tokens.append((match.start() + start, Text, text[start:]))
                text = text_stripped

            # Split trailing close-files into separate tokens
            while text.endswith(")") and text.count("(") < text.count(")"):
                text = text[:-1]
                tokens.insert(0, (match.start() + len(text), IO.CloseFile, ")"))
        if text:
            tokens.insert(0, (0, Text, text))
        return tokens

    tokens = {
        "inputs": [
            (r"{\.?/[^\s(){}]+}", IO.ReadAux),
            (r"<\.?/[^\s(){}]+(?: \(.*\))?>", IO.ReadImage),
            # TODO: reading fonts
            # (/texlive/mt-msb.cfg)<<ot1tt.cmap>> (./tex/document.tex
            # TODO: reading subsetted fonts, the same but with <filename>
        ],
        "root": [
            # These regexes are acting on the output *after* it's been split into
            # sections that end on a start-page or open-file.
            (r"!.+", Generic.Error),
            (r"^(Overfull|Underfull|.*([Ww]arning|ATTENTION)).*", Generic.Warning),
            (START_PAGE_RE, State.StartPage, "page"),
            (OPEN_FILE_RE, IO.OpenFile, "file"),
            (r"\s*\)", IO.CloseFile, "file"),
            include("inputs"),
            (r".+", text_and_close_files),
        ],
        "page": [
            # Page numbers can also look like [1 <./file>] or [1 {./file}]
            include("inputs"),
            (r"\]", State.EndPage, "#pop"),
        ],
        "file": [
            # After seeing "(/filename" or ")", there can be:
            #   * a space and then another open file, or
            #   * a close file, start page, etc., or
            #   * Arbitrary text from package imports, which might end in a close file.
            # The section we're working on is already split at start-pages and
            # open-files, so let's assume anything following is text and close-files.
            (r"\)", IO.CloseFile),
            default("#pop"),
        ],
    }


class Splitter:
    """Utility class for splitting lines of log output."""

    # TODO: roll this into the lexer?

    PAGE_RE = re.compile(LatexLogLexer.START_PAGE_RE)
    FILE_RE = re.compile(LatexLogLexer.OPEN_FILE_RE)

    @staticmethod
    def split_on_re(regex, value):
        """Split a string into sections starting with a given regex."""
        last_index = 0
        for match in regex.finditer(value):
            section = value[last_index : match.start()]
            if section:
                yield section
                last_index = match.start()
        section = value[last_index:]
        if section:
            yield section

    @classmethod
    def split_on_pages(cls, line):
        """Split a line of output into sections starting with a page number."""
        return cls.split_on_re(cls.PAGE_RE, line)

    @classmethod
    def split_on_io(cls, line):
        """Split a line of output into sections starting with a file opening."""
        return cls.split_on_re(cls.FILE_RE, line)

    @classmethod
    def split(cls, line):
        """Split a line of output by start-page or file-open.

        Yields:
            List[str]: sections of the line, each (other than the first) starting with a
            start-page or file-open.
        """
        for section in cls.split_on_pages(line):
            for section_ in cls.split_on_io(section):
                yield section_


def split(line) -> List[str]:
    """Split a line of output by start-page or file-open.

    Returns:
        List[str]: sections of the line, each (other than the first) starting with a
        start-page or file-open.
    """
    return list(Splitter.split(line))


def _lex(line: str, lexer: LatexLogLexer):
    """Lex a single line of output.

    Yields: (tokentype, value)
    """
    for section in Splitter.split(line):
        for token in lexer.get_tokens(section):
            yield token


def lex(line: str, lexer: LatexLogLexer = None) -> List[Tuple[Any, str]]:
    """Lex a single line of output.

    Returns: list of (tokentype, value)
    """
    if not lexer:
        lexer = LatexLogLexer()
    return list(_lex(line, lexer))



"""Tests for formatter module."""
# pylint: disable=redefined-outer-name,redefined-builtin

import blessings
import pytest

from quietex.formatter import AnsiTerminalFormatter, format, quiet_filter
from quietex.lexer import *  # noqa: F403


@pytest.fixture
def term():
    """Construct a blessings Terminal that always produces formatting codes."""
    return blessings.Terminal(kind="xterm-256color", force_styling=True)


def test_formatting_empty():
    """Test formatting an empty list of tokens."""
    formatter = AnsiTerminalFormatter()
    assert format([], formatter) == ""


EXAMPLE_TOKENS = [
    (IO.CloseFile, ")"),
    (Generic.Error, "error"),
    (IO.OpenFile, "(./open.tex"),
    (Text, " text "),
    (State.StartPage, "[1"),
    (State.EndPage, "]"),
    (IO.ReadAux, "{./aux.map}"),
    (IO.ReadImage, "<./image.png>"),
    (Generic.Warning, "warning"),
    # TODO: UI
]


def test_formatting_quiet(term):
    """Test formatting every kind of token in quiet mode."""
    formatter = AnsiTerminalFormatter(term)
    result = format(quiet_filter(EXAMPLE_TOKENS), formatter)
    expected = term.bright_red("error") + " text [1]" + term.yellow("warning")
    assert result == expected


def test_log_quiet(term):
    """Make sure log messages are filtered in quiet mode."""
    formatter = AnsiTerminalFormatter(term)
    result = format(quiet_filter([(UI.Message, "log"), (Text, "text")]), formatter)
    assert result == "text"


def test_formatting_verbose(term):
    """Test formatting every kind of token in verbose mode."""
    formatter = AnsiTerminalFormatter(term)
    result = format(EXAMPLE_TOKENS, formatter)
    expected = (
        term.dim(")")
        + term.bright_red("error")
        + term.dim("(./open.tex")
        + " text [1]"
        + term.dim("{./aux.map}")
        + term.dim("<./image.png>")
        + term.yellow("warning")
    )
    assert result == expected



"""Handle formatting tokens."""
# pylint: disable=redefined-builtin

import blessings
from pygments import format
from pygments.formatter import Formatter

from .lexer import *  # noqa: F403

__all__ = ["AnsiTerminalFormatter", "contains_error", "format", "quiet_filter"]


def contains_error(token_source):
    """Return whether a list of tokens contains an error token."""
    for token_type, _ in token_source:
        if token_type == Generic.Error:
            return True
    return False


def quiet_filter(token_source):
    """Filter a list of tokens to remove IO message for quiet mode."""
    return [
        (token_type, value)
        for (token_type, value) in token_source
        if IO not in (token_type, token_type.parent) and token_type != UI.Message
    ]


class AnsiTerminalFormatter(Formatter):
    """Handle formatting tokens with ANSI styles."""

    def __init__(self, terminal=None):
        super().__init__()
        if terminal is None:
            terminal = blessings.Terminal()

        self.style = {
            IO.CloseFile: terminal.dim,
            IO.OpenFile: terminal.dim,
            IO.ReadAux: terminal.dim,
            IO.ReadImage: terminal.dim,
            Generic.Error: terminal.bright_red,
            Generic.Warning: terminal.yellow,
            UI.Prompt: terminal.red,
            UI.Status: terminal.blue,
            UI.Message: terminal.dim,
        }

    def format_unencoded(self, token_source, outfile):
        """Format a list of tokens.

        Called internally by Formatter.format.
        """
        for token_type, value in token_source:
            assert "\n" not in value
            if token_type in self.style:
                value = self.style[token_type](value)
            outfile.write(value)



"""Tests for command-line interface."""
# pylint: disable=redefined-outer-name,unused-argument
import importlib.resources as pkg_resources
import re
import subprocess
from pathlib import Path
from typing import List

import pytest

import quietex
from quietex.cli import main


def remove_control_sequences(line: str):
    """Remove ANSI controls sequences from a string."""
    # https://stackoverflow.com/a/33925425
    ansi_escape = re.compile(r"(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]")
    return ansi_escape.sub("", line)


def run(cmd: List[str]) -> str:
    """Run a command in a subprocess, check it succeeds, and return its stdout."""
    return subprocess.run(cmd, text=True, check=True, capture_output=True).stdout


def _run_quietex_internal(args: List[str], capsys) -> str:
    """Call the quietex main function and return its stdout.

    Args:
        args: quietex CLI arguments
        capsys: instance of pytest capsys fixture
    """
    main(args)
    return capsys.readouterr().out


@pytest.fixture(params=["script", "module", "internal"])
def run_quietex(request, capsys):
    """Run quietex, repeating the test for each way of running it."""
    if request.param == "script":
        # Run `quietex` on the command line
        yield lambda args: run(["quietex"] + args)
    elif request.param == "module":
        # Run `python -m quietex` on the command line
        yield lambda args: run(["python", "-m", "quietex"] + args)
    else:
        # Run quietex in-process, so it's debuggable and coverage sees it
        yield lambda args: _run_quietex_internal(["quietex"] + args, capsys)


def test_cli_works(run_quietex):
    """Test that the CLI is in the path, and passes through output."""
    output = run_quietex(["--quiet", "echo", "test"])
    assert remove_control_sequences(output).strip() == "test"


@pytest.mark.parametrize(
    "args", ("", "--force", "--verbose --no-bell", "--force --verbose --no-bell")
)
def test_latexmkrc(run_quietex, args):
    """Test that quietex --latexmkrc prints something vaguely correct."""
    cmd = ["--latexmkrc"]
    if args:
        cmd += args.split(" ")
    output = run_quietex(cmd)
    assert re.search(r'\$pdflatex = ".*?quietex .*?\$pdflatex"', output)
    for arg in args.split(" "):
        if arg != "--force":
            assert arg in output


# pylint: disable=invalid-name
installed_only = pytest.mark.skipif(
    Path(quietex.__file__) == Path(__file__).parent.parent / "src/quietex/__init__.py",
    reason="installed package only",
)


@installed_only
def test_sentinel_file_missing():
    """Test that when quietex is installed, the sentinel file is missing.

    This is to make sure we're definitely running tests against the installed version
    (not against the source or an editable install), since the sentinel file doesn't
    make it into the wheel.
    """
    with pytest.raises(FileNotFoundError):
        pkg_resources.read_text(quietex, "sentinel.txt")


def latexmkrc(args="--latekmkrc"):
    """Generate example latexmkrc file."""
    return f"eval `quietex {args}`;"


def document():
    """Generate example LaTeX document."""
    return r"""
    \documentclass{article}
    \begin{document}
    Test
    \end{document}
    """


def test_latexmk(in_temp_dir):
    """Test running latexmkrc with the QuieTeX integration."""
    open("latexmkrc", "w").write(latexmkrc())
    open("main.tex", "w").write(document())
    output = run(["latexmk"])
    assert Path("main.tex").exists()
    assert output.strip().endswith("are up-to-date")

    # No "QuieTeX enabled" message, and no $pdflatex already defined warning
    assert "quietex" not in output.lower()


@pytest.mark.parametrize(
    "quietex_args,cmd_args",
    (("--quiet", ""), ("--quiet", "--quiet"), ("--quiet", "test -- -n --help")),
)
def test_arguments_are_untouched(quietex_args, cmd_args):
    """Test that quietex passes through exactly the right arguments to the command."""
    echo_args = str(Path(__file__).parent / "echo_args")
    cmd = ["quietex"] + quietex_args.split(" ") + [echo_args] + cmd_args.split(" ")
    output = run([a for a in cmd if a])
    assert remove_control_sequences(output).strip() == cmd_args



"""Tests for TerminalFrontend."""
# pylint: disable=protected-access,invalid-name

import re
from typing import Callable, List

import pyte

from quietex.frontend import TerminalFrontend
from test.test_BasicFrontend import (
    EXAMPLE_OUTPUT,
    EXAMPLE_QUIET,
    EXAMPLE_VERBOSE,
    _frontend_integration_test,
)


class AtScreen(pyte.Screen):
    """Pyte screen with the background filled with @s for better testing."""

    @property
    def default_char(self):
        """The background character."""
        return pyte.screens.Char("@")

    def reset(self):
        """Reset the terminal to its initial state."""
        super().reset()
        # Cursor has to have the same character as the background or erasing won't work
        self.cursor.attrs = self.default_char


class FakeTerminalFrontend(TerminalFrontend):
    """TerminalFrontend which outputs to a Pyte emulated terminal instead of stdout."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.screen = AtScreen(80, 24)
        self.stream = pyte.Stream(self.screen)
        self.pre_input_hooks: List[Callable] = []

    def _input(self, raw_prompt):
        self._write(raw_prompt)
        for hook in self.pre_input_hooks:
            hook()
        # Simulate input
        self.stream.feed("\r\n")
        return ""

    def _write(self, raw_value):
        # Swap \n for \r\n
        raw_value = re.sub(r"([^\r])\n", r"\1\r\n", raw_value)
        if raw_value == "\n":
            raw_value = "\r\n"
        # Write to terminal
        self.stream.feed(raw_value)
        return len(raw_value)

    def _get_terminal_width(self):
        return self.screen.columns

    def assert_cursor(self, x, y):  # pylint: disable=invalid-name
        """Assert the cursor is at (`x`, `y`)."""
        assert (self.screen.cursor.x, self.screen.cursor.y) == (x, y)

    def assert_display_like(self, lines, message=None):
        """Make an assertion about what's on the display.

        If `lines` is a string, assert the first line of the display is equal to it.  If
        `lines` is a list of strings, assert each line of the display is equal to the
        corresponding line from the list.  Line endings will be checked.
        """
        if isinstance(lines, str):
            lines = [lines]
        for i, line in enumerate(lines):
            actual = self.screen.display[i]
            expected = (line + self.screen.default_char.data)[: self.screen.columns]

            # Format display line for assertion failure message
            try:
                actual_end_index = actual.index(self.screen.default_char.data)
            except ValueError:
                actual_end_index = self.screen.columns
            actual_to_print = actual[: max(len(expected), actual_end_index)]
            error = f"Failed at line {i}, {repr(actual_to_print)} should be {expected}"
            if message is not None:
                error = f"{message}: {error}"
            assert actual.startswith(expected), error


def test_faketerminalfrontend_write_line():
    r"""Test FakeTerminalFrontend.write handles a line terminated by \n properly."""
    frontend = FakeTerminalFrontend()
    frontend._write("test\n")
    frontend.assert_cursor(0, 1)


def test_faketerminalfrontend_write_blank_line():
    """Test FakeTerminalFrontend.write handles a line followed by a blank line properly."""
    frontend = FakeTerminalFrontend()
    frontend._write("test\n\n")
    frontend.assert_cursor(0, 2)


def test_faketerminalfrontend_write_newline():
    r"""Test FakeTerminalFrontend.write handles a \n on its own properly."""
    frontend = FakeTerminalFrontend()
    frontend._write("test")
    frontend._write("\n")
    frontend.assert_cursor(0, 1)


def test_faketerminalfrontend_write_newlines():
    r"""Test FakeTerminalFrontend.write handles multiple \ns on their own properly."""
    frontend = FakeTerminalFrontend()
    frontend._write("test")
    frontend._write("\n\n")
    frontend.assert_cursor(0, 2)


def test_faketerminalfrontend_wrap_line():
    r"""Test FakeTerminalFrontend.write wraps at 80 characters."""
    frontend = FakeTerminalFrontend()
    frontend._write("x" * 100)
    frontend.assert_display_like(["x" * 80, "x" * 20])


def test_print():
    """Test basic use of print."""
    frontend = FakeTerminalFrontend()
    frontend.print("test")
    frontend.assert_display_like(["test", ""])
    frontend.assert_cursor(0, 1)


def test_print_multiple_lines():
    """Test printing multiple lines (without status)."""
    frontend = FakeTerminalFrontend()
    frontend.print("test1")
    frontend.print("test2")
    frontend.assert_display_like(["test1", "test2", ""])


def test_print_with_status():
    """Test basic use of print with a status line."""
    frontend = FakeTerminalFrontend()
    frontend.print("test1")
    frontend.print("test2 [1]")
    frontend.assert_display_like(["test1", "test2 [1]", "[1]", ""])


def test_print_multiple_lines_with_status():
    """Test the status line stays at the bottom when multiple lines are printed."""
    frontend = FakeTerminalFrontend()
    frontend.print("test1 [1]")
    frontend.print("test2")
    frontend.assert_display_like(["test1 [1]", "[1]", "test2", "[1]", ""])

    frontend.print("test3")
    frontend.assert_display_like(["test1 [1]", "[1]", "test2", "test3", "[1]", ""])


def test_input():
    """Test faking input when no status line has been printed."""
    frontend = FakeTerminalFrontend()
    frontend.print("test")
    frontend.pre_input_hooks = [
        lambda: frontend.assert_display_like(["test", "? ", ""]),
        lambda: frontend.assert_cursor(2, 1),
    ]
    frontend.input("? ")
    frontend.assert_display_like(["test", "? ", ""])

    frontend.print("test2")
    frontend.assert_display_like(["test", "? ", "test2", ""])


def test_input_after_status():
    """Test input prompt prints without status line."""
    frontend = FakeTerminalFrontend()
    frontend.print("test1 [1]")
    frontend.print("test2")
    frontend.pre_input_hooks = [
        lambda: frontend.assert_display_like(["test1 [1]", "[1]", "test2", "? ", ""]),
        lambda: frontend.assert_cursor(2, 3),
    ]
    frontend.input("? ")
    frontend.assert_display_like(["test1 [1]", "[1]", "test2", "? ", ""])

    frontend.print("test3")
    frontend.assert_display_like(
        ["test1 [1]", "[1]", "test2", "? ", "test3", "[1]", ""]
    )


def test_log_clears_status():
    """Test that printing a log message either clears or leaves status correctly."""
    frontend = FakeTerminalFrontend()
    frontend.print("test [1]")
    frontend.log("log")
    frontend.assert_display_like(["test [1]", "[1]", "log", "[1]", ""])
    frontend.log("log2")
    frontend.assert_display_like(["test [1]", "[1]", "log", "log2", "[1]", ""])


def test_verbose():
    """Test printing each type of token in verbose mode."""
    _frontend_integration_test(FakeTerminalFrontend(), EXAMPLE_OUTPUT, EXAMPLE_VERBOSE)


def test_quiet():
    """Test printing each type of token in quiet mode."""
    frontend = FakeTerminalFrontend(quiet=True)
    _frontend_integration_test(frontend, EXAMPLE_OUTPUT, EXAMPLE_QUIET)


def test_line_wrap():
    """Test clearing status bars that are so long they wrap around."""
    frontend = FakeTerminalFrontend()
    msg = (
        "(/a/very/long/filename/which/is/so/long/that/it/wraps/around/onto/the/next"
        "/line/of/the/terminal"
    )
    frontend.print(msg)
    frontend.assert_display_like([msg[:80], msg[80:], msg[:80], msg[80:] + ")"])
    frontend.print("test")
    frontend.print("test2")
    frontend.assert_display_like(
        [
            # Original message
            msg[:80],
            msg[80:],
            # Status bar, since it changed
            msg[:80],
            msg[80:] + ")",
            # Plain text
            "test",
            "test2",
            # Status bar
            msg[:80],
            msg[80:] + ")",
        ]
    )


def test_print_partial_simple():
    """Test printing partial lines with a simple example."""
    frontend = FakeTerminalFrontend()
    frontend.print("Partial", finished=False)
    frontend.assert_display_like("Partial", "")

    frontend.print("Full")
    frontend.assert_display_like("Full", "")


def test_print_partial_complex():
    """Test printing partial lines with a long example including status changes."""
    frontend = FakeTerminalFrontend()
    msg = "(./test.tex [1] (./test2.tex test message [2]))"
    for i in range(1, 4):
        # No valid file yet
        frontend.print(msg[:i], finished=False)
        frontend.assert_display_like([msg[:i], ""], i)
    for i in range(4, 12):
        # Part-way through test.tex
        frontend.print(msg[:i], finished=False)
        frontend.assert_display_like([msg[:i], msg[:i] + ")", ""], i)
    for i in range(12, 14):
        # Part-way through start page
        frontend.print(msg[:i], finished=False)
        frontend.assert_display_like([msg[:i], "(./test.tex)", ""], i)
    for i in range(21, 29):
        # Partway through second file
        frontend.print(msg[:i], finished=False)
        frontend.assert_display_like([msg[:i], msg[12:i] + ")", ""], i)
    for i in range(29, 44):
        # Not up to second page yet
        frontend.print(msg[:i], finished=False)
        frontend.assert_display_like([msg[:i], "[1] (./test2.tex)", ""], i)
    for i in range(44, 45):
        # Not up to close files yet
        frontend.print(msg[:i], finished=False)
        frontend.assert_display_like([msg[:i], "[2] (./test2.tex)", ""], i)
    # Close second file
    frontend.print(msg[:-1], finished=False)
    frontend.assert_display_like([msg[:-1], "[2] (./test.tex)", ""])
    # Close first file
    frontend.print(msg)
    frontend.assert_display_like([msg, "[2]", ""])



"""Main program implementation."""
import argparse
import importlib.resources as pkg_resources
import os
import sys
from typing import List

import pexpect

from ._meta import __version__  # noqa: F401
from .frontend import BasicFrontend, TerminalFrontend  # noqa: F401


def handle_prompt(tty: BasicFrontend, pdflatex: pexpect.spawn):
    """Check if pdflatex has prompted for user input and if so handle it.

    If the "? " prompt is detected, prompt the user for a command (or Ctrl+C or
    Ctrl+D) and pass their response to pdflatex.
    """
    if pdflatex.buffer != "? ":
        # It's not waiting for input
        return
    try:
        prompt = pdflatex.read(2)
    except pexpect.exceptions.TIMEOUT:
        # It's not waiting for input
        return
    while True:
        # pdflatex needs a newline after Ctrl+C, so loop until we get a proper
        # response from the user
        try:
            user_response = tty.input(prompt)
            pdflatex.send(user_response + "\n")
            return
        except EOFError:
            # Ctrl+D at input prompt (pdflatex responds immediately)
            pdflatex.sendcontrol("d")
            return
        except KeyboardInterrupt:
            # Ctrl+C at input prompt (pdflatex responds at end of line)
            pdflatex.sendintr()
            prompt = ""


def run_command(cmd: List[str], **kwargs):
    """Run the command, filtering and colouring its output.

    The command is assumed to be a pdflatex invocation, but other LaTeX compilers
    probably work too.

    Args:
        cmd: Command to run, and its arguments.
        quiet: Whether to completely hide useless output or just dim it.
    """
    # Disable pdflatex line wrap (where possible)
    env = dict(os.environ, max_print_line="1000000000")

    # Run pdflatex and filter/colour output
    pdflatex = pexpect.spawn(cmd[0], cmd[1:], env=env, encoding="utf-8", timeout=0.2)

    tty = TerminalFrontend(**kwargs)
    # tty = BasicFrontend(quiet=quiet)
    tty.log("QuieTeX enabled")

    while True:
        try:
            line = pdflatex.readline()
        except pexpect.exceptions.TIMEOUT:
            # Check if it's waiting for input
            handle_prompt(tty, pdflatex)
            continue
        if line == "":
            # EOF
            break

        # TODO: Page numbers would work better if it parsed the line bit by bit
        tty.print(line.strip("\r\n"))

        # TODO: If you add a 0.1s delay here, it sometimes misses a bit of output at the
        #       end.  Could be related to pexpect/pexpect#120 or
        #       https://pexpect.readthedocs.io/en/stable/commonissues.html#truncated-output-just-before-child-exits  # noqa: B950

    # TODO: Only add newline when necessary
    print()

    pdflatex.close()
    if pdflatex.exitstatus != 0:
        sys.exit(pdflatex.exitstatus)


def print_latexmkrc(cmd, force=False):
    """Print latexmk configuration for using QuieTeX."""
    template = pkg_resources.read_text("quietex", "latexmkrc")
    start, no_force_clause, force_clause, end = template.split("# <split>\n")
    print(start, end="")
    if force:
        print(force_clause % dict(cmd=cmd), end="")
    else:
        print(no_force_clause % dict(cmd=cmd), end="")
    print(end, end="")


def split_argv(argv: List[str]):
    """Split argument list into quietex invocation, quietex args, and command to run."""
    quietex_command, *argv = argv
    if "__main__" in quietex_command:
        # argv[0] is the path to __main__.py when `python -m quietex` runs
        quietex_command = "python -m quietex"
    try:
        # Split argument list before the first non-option argument
        first_arg_index = next(i for i, a in enumerate(argv) if not a.startswith("-"))
        quietex_argv = argv[:first_arg_index]
        cmd = argv[first_arg_index:]
        return quietex_command, quietex_argv, cmd
    except StopIteration:
        return quietex_command, argv, None


def parse_args(argv: List[str], command):
    """Parse quietex command-line arguments."""
    parser = argparse.ArgumentParser(
        description="Filter and colourize LaTeX compilation output."
    )
    parser.add_argument("pdflatex", nargs="?", help="the command to run, e.g. pdflatex")
    parser.add_argument("args", nargs="*", help="the command's arguments")
    parser.add_argument(
        "--latexmkrc",
        action="store_true",
        help="print latexmk settings which enable QuieTeX and also colourize latexmk "
        "output; include `eval quietex --latexmkrc` in latexmkrc to use",
    )
    parser.add_argument(
        "--force",
        "-f",
        action="store_true",
        help="when printing latexmk settings, enable quietex even if $pdflatex is "
        "non-default",
    )
    bell = parser.add_mutually_exclusive_group()
    bell.add_argument(
        "--bell", "-b", action="store_true", help="ring bell on LaTeX error (default)"
    )
    bell.add_argument(
        "--no-bell",
        action="store_false",
        dest="bell",
        help="don't ring bell on LaTeX error",
    )
    quiet = parser.add_mutually_exclusive_group()
    quiet.add_argument(
        "--quiet", "-q", action="store_true", help="hide I/O messages (default)"
    )
    quiet.add_argument(
        "--verbose",
        "-v",
        action="store_false",
        dest="quiet",
        help="just dim I/O messages, and also add information messages from QuieTeX",
    )
    parser.set_defaults(quiet=True, bell=True)
    args = parser.parse_args(argv)

    # Easier to check these manually than to explain it to argparse
    if args.latexmkrc and command:
        print("Use --latexmkrc or command, not both\n")
    elif args.force and command:
        print("--force is only valid with --latexmkrc\n")
    elif not args.latexmkrc and not command:
        pass
    else:
        return args
    # Hit one of the error cases
    parser.print_help()
    sys.exit(1)


def main(argv=None):
    """Handle command line arguments."""
    if argv is None:
        argv = sys.argv

    quietex_command, quietex_argv, command = split_argv(argv)
    args = parse_args(quietex_argv, command)

    if args.latexmkrc:
        if not args.quiet:
            quietex_command += " --verbose"
        if not args.bell:
            quietex_command += " --no-bell"
        print_latexmkrc(quietex_command, args.force)
    else:
        run_command(command, quiet=args.quiet, bell_on_error=args.bell)



"""Manage application state and status bar."""

from typing import List, Optional

import attr
from attr import attrs

from .lexer import IO, State


@attrs(auto_attribs=True, frozen=True)
class AppState:
    """Manage application state and status bar."""

    current_page: Optional[int] = None
    file_stack: List[str] = attr.Factory(list)

    @property
    def current_file(self):
        """The file currently being processed."""
        if self.file_stack:
            return self.file_stack[-1]
        return None

    def update(self, token_source):
        """Update current file and page based on tokens to print."""
        next_page = self.current_page
        next_stack = self.file_stack.copy()
        for token_type, value in token_source:
            if token_type == State.StartPage:
                next_page = int(value.strip("[] "))
            elif token_type == IO.OpenFile:
                next_stack.append(value.strip("("))
            elif token_type == IO.CloseFile:
                try:
                    next_stack.pop()
                except IndexError:
                    pass
        return AppState(next_page, next_stack)

    def format_status(self):
        """Return the current status bar as a string, and reset dirtiness."""
        status = ""
        if self.current_page:
            status += f"[{self.current_page}]"
        if self.current_file:
            status += f" ({self.current_file})"
            status = status.strip(" ")
        return status



"""QuieTeX setup."""
# pylint: disable=invalid-name
from setuptools import setup

with open("Readme.md") as f:
    readme_lines = f.readlines()
    long_description = "\n".join(readme_lines[: readme_lines.index("## Development\n")])

meta = {}  # type: ignore
with open("src/quietex/_meta.py") as fp:
    exec(fp.read(), meta)  # pylint: disable=exec-used

tests_require = ["pyte", "pytest", "pytest-cov"]


setup(
    name="quietex",
    version=meta["__version__"],
    description="Filter output of pdflatex.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mje-nz/quietex",
    author="Matthew Edwards",
    license="BSD 3-Clause",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Text Processing :: Markup :: LaTeX",
    ],
    packages=["quietex"],
    package_dir={"": "src"},
    package_data={"quietex": ["latexmkrc"]},
    entry_points={"console_scripts": ["quietex=quietex.cli:main"]},
    python_requires=">=3.7",
    install_requires=["attrs", "blessings", "pexpect", "pygments"],
    extras_require={"test": tests_require},
    tests_require=tests_require,
    zip_safe=True,
)



"""Input and output handling.

Author: Matthew Edwards
Date: August 2019
"""
import shutil
import sys
from typing import Any, List, Tuple

# pylint: disable=redefined-builtin
from .formatter import AnsiTerminalFormatter, contains_error, format, quiet_filter
from .lexer import UI, LatexLogLexer, lex
from .status import AppState


class BasicFrontend:
    """Handle input and output with optional colour but no cursor movement."""

    def __init__(self, quiet=False, bell_on_error=False):
        self.quiet = quiet
        self.bell_on_error = bell_on_error
        self.state = AppState()
        self.lexer = LatexLogLexer()
        self.formatter = AnsiTerminalFormatter()

    def _input(self, raw_prompt):
        """Display a prompt and return the user's input."""
        return input(raw_prompt)

    def input(self, prompt):
        """Display a prompt (with highlighting) and return the user's input."""
        return self._input(format([(UI.Prompt, prompt)], self.formatter))

    def _write(self, raw_value):
        """Write directly to the underlying output.

        Returns:
            int: number of characters written.
        """
        sys.stdout.write(raw_value)
        sys.stdout.flush()  # Make sure lines without newlines are printed

    def _print_tokens(self, tokens: List[Tuple[Any, str]], end="\n"):
        """Highlight and print a list of tokens, updating app state if necessary."""
        if not tokens:
            # Make sure blank lines get printed
            self._write(end)
            return 0
        if self.quiet:
            tokens = quiet_filter(tokens)
        if self.bell_on_error and contains_error(tokens):
            end += "\a"
        if tokens:
            # Skip line if it's now empty to avoid lone newline
            self._write(format(tokens, self.formatter) + end)
        length = len("".join(value for (_, value) in tokens))
        return length

    def _print_status(self, status, end="\n"):
        self._print_tokens([(UI.Status, status)], end=end)

    def log(self, message):
        """Print a log message."""
        self._print_tokens([(UI.Message, message)])

    def print(self, value: str, finished=True):
        """Lex, highlight, and print a line of LaTeX compiler output.

        Args:
            value: Line of LaTeX compiler output to print.
            finished: Whether the line is complete, i.e., ended in a newline.
        """
        if not finished:
            return
        tokens = lex(value, self.lexer)
        self._print_tokens(tokens)
        new_state = self.state.update(tokens)
        if new_state != self.state:
            self._print_status(new_state.format_status())
            self.state = new_state


class TerminalFrontend(BasicFrontend):
    """Handle input and output with cursor movement and optional colour."""

    CURSOR_UP = "\x1b[A"
    CURSOR_TO_START = "\x1b[G"
    DELETE_WHOLE_LINE = "\x1b[2K"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.keep_last_status = False
        self.last_line_length = None

    def _get_terminal_width(self):
        return shutil.get_terminal_size().columns

    def _clear_status(self):
        """Clear the status bar (unless marked keep) and last line (if not finished)."""
        if self.keep_last_status:
            # Finish status line first
            self._write("\n")
            self.keep_last_status = False
        else:
            last_status_length = len(self.state.format_status())
            # Status bar doesn't end in a newlineClear status bar first
            self._write(self.CURSOR_TO_START + self.DELETE_WHOLE_LINE)
            # Clear previous lines if the status bar is more than one line long
            for _ in range(last_status_length // self._get_terminal_width()):
                self._write(self.CURSOR_UP + self.DELETE_WHOLE_LINE)

            # Also clear last line printed if it wasn't finished
            if self.last_line_length:
                for _ in range(self.last_line_length // self._get_terminal_width() + 1):
                    self._write(self.CURSOR_UP + self.DELETE_WHOLE_LINE)

    def input(self, *args, **kwargs):  # pylint: disable=arguments-differ
        """Display a prompt with the given style and return the user's input.

        The status line will be cleared if it is displayed, and it will not be
        re-printed until after the next line of output.
        """
        self._clear_status()
        return super().input(*args, **kwargs)

    def log(self, message):
        """Print a log message."""
        self._clear_status()
        super().log(message)
        self._print_status(self.state.format_status(), end="")

    def print(self, value="", finished=True):  # pylint: disable=arguments-differ
        """Lex, highlight, and print a line of LaTeX compiler output."""
        self._clear_status()
        tokens = lex(value, self.lexer)
        length = self._print_tokens(tokens)
        new_state = self.state.update(tokens)
        self._print_status(new_state.format_status(), end="")
        if finished:
            self.last_line_length = None
        else:
            self.last_line_length = length
        if finished and new_state != self.state:
            self.state = new_state
            self.keep_last_status = True



"""Tests for BasicFrontend."""
# pylint: disable=invalid-name

import io
from typing import List

from quietex.frontend import BasicFrontend


class StringBasicFrontend(BasicFrontend):
    """BasicIO which records its outputs in a StringIO."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._output = io.StringIO()

    def _write(self, raw_value):
        return self._output.write(raw_value)

    def _input(self, raw_prompt):
        self._write(raw_prompt + "\n")
        return ""

    @property
    def output(self):
        """Everything printed so far, as a string."""
        return self._output.getvalue()

    def assert_display_like(self, lines):
        r"""Make an assertion about what's on the display.

        If `lines` is a string, assert that the output so far is equal to it.  If
        `lines` is a list of strings, assert that the output so far is equal to those
        strings joined with \n.

        Roughly compatible with test_TerminalFrontend.FakeTerminalFrontend.
        """
        if isinstance(lines, str):
            lines = [lines]
        assert self.output == "\n".join(lines)


def test_simple():
    """Print some plain text and check that the status bar is not printed."""
    frontend = StringBasicFrontend()
    frontend.print("Test")
    assert frontend.output == "Test\n"


def test_state():
    """Check that printing StartPage and OpenFile triggers the status bar."""
    frontend = StringBasicFrontend()
    frontend.print("(./test.tex [1]")
    assert frontend.output == "(./test.tex [1]\n[1] (./test.tex)\n"


def test_log_newline():
    """Test that printing a log message puts newlines in the right places."""
    frontend = StringBasicFrontend()
    frontend.print("test [1]")
    frontend.log("log")
    assert frontend.output == "test [1]\n[1]\nlog\n"
    frontend.log("log2")
    assert frontend.output == "test [1]\n[1]\nlog\nlog2\n"


EXAMPLE_OUTPUT = [
    "(./open.tex ",
    "error [1]",
    "test",
    "",
    "{./aux.map}",
    "<./image.png>",
    "warning [2]",
    ")",
]
EXAMPLE_VERBOSE = [
    "(./open.tex ",
    "(./open.tex)",
    "error [1]",
    "[1] (./open.tex)",
    "test",
    "",
    "{./aux.map}",
    "<./image.png>",
    "warning [2]",
    "[2] (./open.tex)",
    ")",
    "[2]",
    "Log message",
    "? ",
    "",
]
EXAMPLE_QUIET = [
    " ",
    "(./open.tex)",
    "error [1]",
    "[1] (./open.tex)",
    "test",
    "",
    "warning [2]",
    "[2] (./open.tex)",
    "[2]",
    "? ",
    "",
]


def _frontend_integration_test(frontend, output: List[str], expected: List[str]):
    for line in output:
        frontend.print(line)
    frontend.log("Log message")
    frontend.input("? ")
    frontend.assert_display_like(expected)


def test_verbose():
    """Test printing each type of token in verbose mode."""
    _frontend_integration_test(StringBasicFrontend(), EXAMPLE_OUTPUT, EXAMPLE_VERBOSE)


def test_quiet():
    """Test printing each type of token in quiet mode."""
    frontend = StringBasicFrontend(quiet=True)
    _frontend_integration_test(frontend, EXAMPLE_OUTPUT, EXAMPLE_QUIET)


def test_error_bell():
    """Test that printing an error adds a bell character."""
    frontend = StringBasicFrontend(bell_on_error=True)
    frontend.print("! Generic error")
    assert "\a" in frontend.output


def test_print_partial():
    """Test that trying to print a partial line does nothing."""
    frontend = StringBasicFrontend()
    frontend.print("test", finished=False)
    assert frontend.output == ""



"""Tests for lexer module."""
from pathlib import Path
from typing import Any, List, Tuple

import pytest

from quietex.lexer import IO, Generic, State, Text, lex, split


def test_split():
    """Test splitting lines by start-page and open-file."""
    msg = [
        r"(/usr/local/texlive/2019/texmf-dist/tex/latex/lm/ot1lmtt.fd) ",
        "[1{/usr/local/texlive/2019/texmf-var/fonts/map/pdftex/updmap/pdftex.map}] ",
        "[2] ",
        "(./Thesis.toc) ",
        "[3]",
    ]
    # TODO: more thorough test cases
    assert list(split("".join(msg))) == msg


def combine_text(tokens):
    """Combine consecutive Text tokens."""
    # TODO: TokenMergeFilter?
    out: List[Tuple[Any, str]] = []
    for (tokentype, value) in tokens:
        if tokentype == Text and out and out[-1][0] == Text:
            out[-1] = (tokentype, out[-1][1] + value)
        else:
            out.append((tokentype, value))
    return out


def test_combine_text():
    """Test combine_text helper function."""
    tokens = [(Text, "a"), (Text, "b"), (IO, "c"), ("Text", "d")]
    assert combine_text(tokens) == [(Text, "ab"), (IO, "c"), ("Text", "d")]


# Output from a simple test document:
# This is pdfTeX, Version 3.14159265-2.6-1.40.20 (TeX Live 2019) (preloaded format=pdflatex)  # noqa: B950
#  restricted \write18 enabled.
# entering extended mode
# (./test.tex
# LaTeX2e <2018-12-01>
# (/usr/local/texlive/2019/texmf-dist/tex/latex/base/article.cls
# Document Class: article 2018/09/03 v1.4i Standard LaTeX document class
# (/usr/local/texlive/2019/texmf-dist/tex/latex/base/size10.clo)) (./test.aux)
# (./include file.tex) [1{/usr/local/texlive/2019/texmf-var/fonts/map/pdftex/updmap/pdftex.map}] (./test.aux) )</usr/local/texlive/2019/texmf-dist/fonts/type1/public/amsfonts/cm/cmr10.pfb>  # noqa: B950
# Output written on test.pdf (1 page, 12659 bytes).
# Transcript written on test.log.


@pytest.mark.parametrize(
    "msg",
    (
        "This is pdfTeX, Version 3.14159265-2.6-1.40.20 (TeX Live 2019) (preloaded "
        "format=pdflatex)",
        r" restricted \write18 enabled.",
        "LaTeX2e <2018-12-01>",
        "Document Class: article 2018/09/03 v1.4i Standard LaTeX document class",
        "Output written on test.pdf (1 page, 12659 bytes).",
        "Transcript written on test.log.",
    ),
)
def test_lex_text(msg):
    """Test lexing some uninteresting log messages."""
    assert combine_text(lex(msg)) == [(Text, msg)]


@pytest.mark.parametrize(
    "msg",
    ("(./test.tex", "(/usr/local/texlive/2019/texmf-dist/tex/latex/base/article.cls"),
)
def test_lex_file_opening_simple(msg):
    """Test lexing some simple file open messages."""
    assert lex(msg) == [(IO.OpenFile, msg.strip())]


@pytest.mark.parametrize("msg", (")", " )"))
def test_lex_file_closing_simple(msg):
    """Test lexing the simplest file close message."""
    assert lex(msg) == [(IO.CloseFile, msg)]


def test_lex_open_close_file():
    """Test lexing an open file message followed by a close file message."""
    assert lex("(./test.tex)") == [(IO.OpenFile, "(./test.tex"), (IO.CloseFile, ")")]


def test_lex_files_complex():
    """Test lexing many open/close file messages on one line."""
    msg = ")) (/usr/a) (/usr/b (/usr/c (/usr/d)) (/usr/e) (/usr/f)) (/usr/g (/usr/h))"
    assert lex(msg) == [
        (IO.CloseFile, ")"),
        (IO.CloseFile, ")"),
        (Text, " "),
        (IO.OpenFile, "(/usr/a"),
        (IO.CloseFile, ")"),
        (Text, " "),
        (IO.OpenFile, "(/usr/b"),
        (Text, " "),
        (IO.OpenFile, "(/usr/c"),
        (Text, " "),
        (IO.OpenFile, "(/usr/d"),
        (IO.CloseFile, ")"),
        (IO.CloseFile, ")"),
        (Text, " "),
        (IO.OpenFile, "(/usr/e"),
        (IO.CloseFile, ")"),
        (Text, " "),
        (IO.OpenFile, "(/usr/f"),
        (IO.CloseFile, ")"),
        (IO.CloseFile, ")"),
        (Text, " "),
        (IO.OpenFile, "(/usr/g"),
        (Text, " "),
        (IO.OpenFile, "(/usr/h"),
        (IO.CloseFile, ")"),
        (IO.CloseFile, ")"),
    ]


def test_lex_message_inside_files():
    """Test lexing a message hidden amongst open/close file messages."""
    msg = [
        "(/usr/local/texlive/2019/texmf-dist/tex/latex/fp/fp.sty",
        " `Fixed Point Package', Version 0.8, April 2, 1995 (C) Michael Mehlich ",
        "(/usr/local/texlive/2019/texmf-dist/tex/latex/fp/defpattern.sty",
        ")",
    ]
    assert combine_text(lex("".join(msg))) == [
        (IO.OpenFile, msg[0]),
        (Text, msg[1]),
        (IO.OpenFile, msg[2]),
        (IO.CloseFile, msg[3]),
    ]


@pytest.mark.parametrize(
    "msg",
    (
        (
            (
                "(/usr/local/texlive/2019/texmf-dist/tex/latex/fp/fp.sty",
                " `Fixed Point Package', Version 0.8, April 2, 1995 (C) Michael Mehlich",
                ")",
                " ",
                "(/usr/local/texlive/2019/texmf-dist/tex/latex/fp/defpattern.sty",
                ")",
            ),
            (
                "(/usr/local/texlive/2019/texmf-dist/tex/latex/lineno/lineno.sty",
                " Style option: `fancyvrb' v3.2a <2019/01/15> (tvz)",
                ")",
                " ",
                "(/usr/local/texlive/2019/texmf-dist/tex/latex/upquote/upquote.sty",
                ")",
            ),
        )
    ),
)
def test_lex_message_followed_by_close(msg):
    """Test lexing a message hidden amongst open/close file messages.

    This is basically the last one but with a CloseFile after the message.
    """
    assert combine_text(lex("".join(msg))) == [
        (IO.OpenFile, msg[0]),
        (Text, msg[1]),
        (IO.CloseFile, ")"),
        (Text, " "),
        (IO.OpenFile, msg[4]),
        (IO.CloseFile, msg[5]),
    ]


def test_lex_message_inside_files_fancyvrb():
    """Test lexing the message from loading fancyvrb, which has brackets in it (!)."""
    msg = [
        "(/usr/local/texlive/2019/texmf-dist/tex/latex/fancyvrb/fancyvrb.sty",
        " Style option: `fancyvrb' v3.2a <2019/01/15> (tvz)",
        ")",
        " ",
        "(/usr/local/texlive/2019/texmf-dist/tex/latex/upquote/upquote.sty",
    ]
    assert lex("".join(msg)) == [
        (IO.OpenFile, msg[0]),
        (Text, msg[1]),
        (IO.CloseFile, msg[2]),
        (Text, msg[3]),
        (IO.OpenFile, msg[4]),
    ]


def test_lex_message_inside_files_fancyvrb_minimal():
    """Test lexing the message from loading fancyvrb, which has brackets in it (!).

    Make sure it works without the next open message.
    """
    msg = [
        "(/usr/local/texlive/2019/texmf-dist/tex/latex/fancyvrb/fancyvrb.sty",
        " Style option: `fancyvrb' v3.2a <2019/01/15> (tvz)",
        ")",
    ]
    assert lex("".join(msg)) == [
        (IO.OpenFile, msg[0]),
        (Text, msg[1]),
        (IO.CloseFile, msg[2]),
    ]


@pytest.mark.parametrize(
    "msg",
    (
        "<./img/test.jpg>",
        "<./img/test.jpg>",
        "<./img/test.png>",
        "<./img/test.png (PNG copy)>",
        "</usr/local/texlive/2019/texmf-dist/fonts/type1/public/lm/lmbx10.pfb>",
        # TODO: Names with spaces
    ),
)
def test_lex_read_image(msg):
    """Test lexing some read images messages"""
    assert lex(msg) == [(IO.ReadImage, msg)]


def test_lex_read_image_in_warning():
    """Test lexing a read image inside a page number at the end of a warning."""
    line = r"Underfull \vbox (badness 7963) has occurred while \output is active [2 <./test.png (PNG copy)>]"  # noqa: B950
    assert lex(line) == [
        (
            Generic.Warning,
            r"Underfull \vbox (badness 7963) has occurred while \output is active ",
        ),
        (State.StartPage, "[2 "),
        (IO.ReadImage, "<./test.png (PNG copy)>"),
        (State.EndPage, "]"),
    ]


@pytest.mark.parametrize(
    "msg",
    (
        "{/usr/local/texlive/2019/texmf-dist/fonts/enc/dvips/lm/lm-mathit.enc}",
        "{/usr/local/texlive/2019/texmf-var/fonts/map/pdftex/updmap/pdftex.map}",
        # TODO: Names with spaces
    ),
)
def test_lex_read_aux(msg):
    """Test lexing some read aux messages"""
    assert lex(msg) == [(IO.ReadAux, msg)]


def test_lex_read_aux_in_page():
    """Test lexing a read aux inside a page number."""
    line = "[1{/usr/local/texlive/2019/texmf-var/fonts/map/pdftex/updmap/pdftex.map}]"
    assert lex(line) == [
        (State.StartPage, "[1"),
        (
            IO.ReadAux,
            "{/usr/local/texlive/2019/texmf-var/fonts/map/pdftex/updmap/pdftex.map}",
        ),
        (State.EndPage, "]"),
    ]


# New output from a simple test document with an error:
# ! Undefined control sequence.
# l.6 \badcommand


def test_lex_error():
    """Test lexing an error message."""
    msg = "! Undefined control sequence."
    assert lex(msg) == [(Generic.Error, msg)]


# A few warnings from a more complicated document
# Underfull \vbox (badness 10000) has occurred while \output is active [19]
# Overfull \hbox (2.79605pt too wide) detected at line 132
# LaTeX Warning: Marginpar on page 7 moved.
# Package natbib Warning: Citation `Siebert2009' on page 17 undefined on input line 22.
# Package microtype Warning: \nonfrenchspacing is active. Adjustment of
# (microtype)                interword spacing will disable it. You might want
# (microtype)                to add `\microtypecontext{spacing=nonfrench}'
# (microtype)                to your preamble.
# pdfTeX warning (dest): name{Hfootnote.2} has been referenced but does not exist, replaced by a fixed one  # noqa: B950


# TODO: Find examples for this old code
#     if "warning" in line or "missing" in line:
#         return True
#     if "undefined" in line and "i'll" not in line:
#         # Check for "I'll forget about whatever was undefined." in help text
#         return True


@pytest.mark.parametrize(
    "msg",
    (
        r"Underfull \vbox (badness 10000) has occurred while \output is active",
        r"Overfull \hbox (2.79605pt too wide) detected at line 132",
        "LaTeX Warning: Marginpar on page 7 moved.",
        "Package natbib Warning: Citation `Siebert2009' on page 17 undefined on input line 22.",  # noqa: B950
        "pdfTeX warning (dest): name{Hfootnote.2} has been referenced but does not exist, replaced by a fixed one",  # noqa: B950
        "** ATTENTION: Overriding command lockouts (line 45).",
    ),
)
def test_lex_simple_warnings(msg):
    """Test lexing simple warning messages."""
    assert lex(msg) == [(Generic.Warning, msg)]


# A few examples of page numbers and false positives from a complicated document
# [Loading MPS to PDF converter (version 2006.09.02).]
# ) (/usr/local/texlive/2019/texmf-dist/tex/latex/oberdiek/epstopdf-base.sty (/usr/local/texlive/2019/texmf-dist/tex/latex/oberdiek/grfext.sty) (/usr/local/texlive/2019/texmf-dist/tex/latex/latexconfig/epstopdf-sys.cfg)) (/usr/local/texlive/2019/texmf-dist/tex/latex/oberdiek/pdflscape.sty (/usr/local/texlive/2019/texmf-dist/tex/latex/graphics/lscape.sty)) ABD: EveryShipout initializing macros (/usr/local/texlive/2019/texmf-dist/tex/latex/translations/translations-basic-dictionary-english.trsl) (./pythontex-files-Thesis/Thesis.pytxmcr) (./pythontex-files-Thesis/Thesis.pytxpyg) (/usr/local/texlive/2019/texmf-dist/tex/latex/lm/ot1lmr.fd) (/usr/local/texlive/2019/texmf-dist/tex/latex/lm/omllmm.fd) (/usr/local/texlive/2019/texmf-dist/tex/latex/lm/omslmsy.fd) (/usr/local/texlive/2019/texmf-dist/tex/latex/lm/omxlmex.fd) (/usr/local/texlive/2019/texmf-dist/tex/latex/amsfonts/umsa.fd) (/usr/local/texlive/2019/texmf-dist/tex/latex/microtype/mt-msa.cfg) (/usr/local/texlive/2019/texmf-dist/tex/latex/amsfonts/umsb.fd) (/usr/local/texlive/2019/texmf-dist/tex/latex/microtype/mt-msb.cfg) (/usr/local/texlive/2019/texmf-dist/tex/latex/lm/ot1lmss.fd) (/usr/local/texlive/2019/texmf-dist/tex/latex/lm/ot1lmtt.fd) [1{/usr/local/texlive/2019/texmf-var/fonts/map/pdftex/updmap/pdftex.map}] [2] (./Thesis.toc) (./Thesis.gls-abr [3] [4]) (./Thesis.sls [5] [6]) [7] [8] (./Thesis.tdo [9]) [10] (./chapters/fiducial_marker_system/fiducial_marker_system.tex  # noqa: B950
# [17 <./chapters/checkerboards//img/checkerboard.pdf>]
# [18]
#  [][][][]
# Underfull \vbox (badness 7963) has occurred while \output is active [2 <./chapters/fiducial_marker_system//img/basic_marker.png (PNG copy)>]  # noqa: B950
# Underfull \vbox (badness 10000) has occurred while \output is active [19]
# \OT1/lmr/bx/n/10.95 H \OT1/lmr/m/n/10.95 = [] [] \OML/lmm/m/it/10.95 :


def test_lex_page_number_simple():
    """Test lexing a simple page number."""
    assert lex("[1]") == [(State.StartPage, "[1"), (State.EndPage, "]")]


@pytest.mark.parametrize(
    "msg",
    (
        "(./Thesis.gls-abr [1])",
        ") (/usr/local/texlive/2019/texmf-dist/tex/latex/lm/ot1lmtt.fd) [1]",
        "[1 <./chapters/fiducial_marker_system//img/basic_marker.png (PNG copy)>]",
        "[1{/usr/local/texlive/2019/texmf-var/fonts/map/pdftex/updmap/pdftex.map}]",
        r"Underfull \vbox (badness 10000) has occurred while \output is active [1]",
        # TODO: [1{/usr/local/texlive/2019/texmf-var/fonts/map/pdftex/updmap/pdftex.map}] [2] (./Thesis.toc)  # noqa: B950
    ),
)
def test_lex_page_number_complex(msg):
    """Test lexing page numbers on the same line as other messages."""
    tokens = lex(msg)
    page_tokens = [(t, v) for (t, v) in tokens if t == State.StartPage]
    assert len(page_tokens) == 1
    assert page_tokens[0][1].strip("[] ") == "1"


def test_lex_page_number_complex_1():
    """Test lexing a page number inside an open/close file."""
    assert lex("(./Thesis.gls-abr [1])") == [
        (IO.OpenFile, "(./Thesis.gls-abr"),
        (Text, " "),
        (State.StartPage, "[1"),
        (State.EndPage, "]"),
        (IO.CloseFile, ")"),
    ]


def test_lex_page_number_complex_2():
    """Test lexing a page number at the end of a warning."""
    msg = r"Underfull \vbox (badness 10000) has occurred while \output is active [1]"
    assert lex(msg) == [
        (Generic.Warning, msg[:-3]),
        (State.StartPage, "[1"),
        (State.EndPage, "]"),
    ]


@pytest.mark.parametrize(
    "msg", ("[Loading MPS to PDF converter (version 2006.09.02).]", " [][][][]")
)
def test_lex_page_number_false_positives(msg):
    """Test lexing things that look like page numbers but aren't."""
    assert lex(msg) == [(Text, msg)]


def test_round_trip():
    """Test nothing is lost when lexing a typical log."""
    log = open(Path(__file__).parent / "thesis.log").read()
    for line in log.splitlines():
        assert line == "".join(token[1] for token in lex(line))


# TODO: handle errors inside <read image> e.g.
ERROR_IN_READ_IMAGE = r"""
<./chapters/corners//img/2018_09_17_ur5_corners_correlation_x.pdf

pdfTeX warning: pdflatex (file ./chapters/corners//img/2018_09_17_ur5_corners_c
orrelation_x.pdf): PDF inclusion: multiple pdfs with page group included in a s
ingle page
>"""


# TODO: I'm guessing I intended to write tests for these

# Underfull vbox warning followed by page numbers and noise
EDGE_CASE_1 = r"""Underfull \vbox (badness 4378) has occurred while \output is active [7]
 [8] [9 <./chapters/fiducial_marker_system//img/bad_marker_closeup.jpg> <./chapters/
 fiducial_marker_system//img/histogram_matching_graph.png>] [10 <./chapters/fiducial_
 marker_system//img/histogram_matching_image.png (PNG copy)> <./chapters/fiducial_marker
 _system//img/parameters_of_interest_distance.png> <./chapters/fiducial_marker_system//
 img/parameters_of_interest_size.png>] [11 <./chapters/fiducial_marker_system//img/
 sinusoid_marker5_frame000.png (PNG copy)> <./chapters/fiducial_marker_system//img/
 sinusoid_marker6_frame000.png (PNG copy)> <./chapters/fiducial_marker_system//img/
 sinusoid_marker5_frame000_cropped.png (PNG copy)> <./chapters/fiducial_marker_system//
 img/sinusoid_marker6_frame000_cropped.png (PNG copy)> <./chapters/fiducial_marker_
 system//img/sinusoid_marker5_stds.png> <./chapters/fiducial_marker_system//img/sinusoid
 _marker6_stds.png>] [12 <./chapters/fiducial_marker_system//img/sinusoid_marker5_
 stdtrend.png> <./chapters/fiducial_marker_system//img/sinusoid_marker6_stdtrend.png>])
 (./chapters/checkerboards/checkerboards.tex [13] [14 <./chapters/fiducial_marker_system
 //img/marker_parameter_histogram_first.png>] [15 <./chapters/fiducial_marker_system//
 img/marker_parameter_histogram_second.png>] [16]""".replace(
    "\n", ""
)

# Output from package inside noise
EDGE_CASE_2 = """) (/usr/local/texlive/2019/texmf-dist/tex/latex/fnpct/fnpct.sty (/usr/
local/texlive/2019/texmf-dist/tex/latex/koma-script/scrlfile.sty) (/usr/local/texlive/
2019/texmf-dist/tex/latex/filehook/filehook-scrlfile.sty) (/usr/local/texlive/2019/
texmf-dist/tex/latex/translations/translations.sty)) (/usr/local/texlive/2019/texmf-dist
/tex/latex/footmisc/footmisc.sty) (/usr/local/texlive/2019/texmf-dist/tex/latex/fnbreak
/fnbreak.sty) (/usr/local/texlive/2019/texmf-dist/tex/latex/todonotes/todonotes.sty
(/usr/local/texlive/2019/texmf-dist/tex/latex/pgf/frontendlayer/tikz.sty (/usr/local/
texlive/2019/texmf-dist/tex/latex/pgf/utilities/pgffor.sty (/usr/local/texlive/2019/
texmf-dist/tex/latex/pgf/utilities/pgfkeys.sty (/usr/local/texlive/2019/texmf-dist/tex/
generic/pgf/utilities/pgfkeys.code.tex)) (/usr/local/texlive/2019/texmf-dist/tex/latex/
pgf/math/pgfmath.sty (/usr/local/texlive/2019/texmf-dist/tex/generic/pgf/math/pgfmath.
code.tex)) (/usr/local/texlive/2019/texmf-dist/tex/generic/pgf/utilities/pgffor.code.tex
(/usr/local/texlive/2019/texmf-dist/tex/generic/pgf/math/pgfmath.code.tex))) (/usr/
local/texlive/2019/texmf-dist/tex/generic/pgf/frontendlayer/tikz/tikz.code.tex (/usr/
local/texlive/2019/texmf-dist/tex/generic/pgf/libraries/pgflibraryplothandlers.code.tex)
(/usr/local/texlive/2019/texmf-dist/tex/generic/pgf/modules/pgfmodulematrix.code.tex)
(/usr/local/texlive/2019/texmf-dist/tex/generic/pgf/frontendlayer/tikz/libraries/
tikzlibrarytopaths.code.tex))) (/usr/local/texlive/2019/texmf-dist/tex/generic/pgf/
frontendlayer/tikz/libraries/tikzlibrarypositioning.code.tex) (/usr/local/texlive/2019/
texmf-dist/tex/generic/pgf/frontendlayer/tikz/libraries/tikzlibraryshadows.code.tex
(/usr/local/texlive/2019/texmf-dist/tex/generic/pgf/frontendlayer/tikz/libraries/
tikzlibraryfadings.code.tex (/usr/local/texlive/2019/texmf-dist/tex/generic/pgf/
libraries/pgflibraryfadings.code.tex)))) (/usr/local/texlive/2019/texmf-dist/tex/latex/
cleveref/cleveref.sty) (/usr/local/texlive/2019/texmf-dist/tex/latex/subfiles/subfiles.
sty (/usr/local/texlive/2019/texmf-dist/tex/latex/tools/verbatim.sty))) (./mph/maths.
tex) (./commands.tex) (./acronyms.tex) (./symbols.tex) (/usr/local/texlive/2019/
texmf-dist/tex/latex/pythontex/pythontex.sty (/usr/local/texlive/2019/texmf-dist/tex/
latex/fvextra/fvextra.sty
(/usr/local/texlive/2019/texmf-dist/tex/latex/fancyvrb/fancyvrb.sty
Style option: `fancyvrb' v3.2a <2019/01/15> (tvz))
(/usr/local/texlive/2019/texmf-dist/tex/latex/upquote/upquote.sty) (/usr/local/texlive/
2019/texmf-dist/tex/latex/lineno/lineno.sty)) (/usr/local/texlive/2019/texmf-dist/tex/
generic/xstring/xstring.sty (/usr/local/texlive/2019/texmf-dist/tex/generic/xstring/
xstring.tex)) (/usr/local/texlive/2019/texmf-dist/tex/latex/pgfopts/pgfopts.sty) (/usr/
local/texlive/2019/texmf-dist/tex/latex/caption/newfloat.sty)) (/usr/local/texlive/
texmf-local/tex/latex/pythontexfigures/pythontexfigures.sty) (./Thesis.aux) (/usr/local/
texlive/2019/texmf-dist/tex/latex/base/ts1cmr.fd) (/usr/local/texlive/2019/texmf-dist/
tex/latex/lm/t1lmr.fd) (/usr/local/texlive/2019/texmf-dist/tex/latex/microtype/mt-cmr.
cfg)""".replace(
    "\n", ""
)

# Package warning with continuation, followed by message on its own line
EDGE_CASE_3 = """Package caption Warning: Unsupported document class (or package) detected,
(caption)                usage of the caption package is not recommended.
See the caption package documentation for explanation.

"""

# File with space in name
EDGE_CASE_4 = """(./include file.tex) [1{/usr/local/texlive/2019/texmf-var/fonts/map/
pdftex/updmap/pdftex.map}] (./test.aux) )</usr/local/texlive/2019/texmf-dist/fonts/
type1/public/amsfonts/cm/cmr10.pfb>""".replace(
    "\n", ""
)


# TODO: get the noise in this to lex
"""(see the transcript file for additional information)
{/usr/local/texlive/2019/texmf-dist/fonts/enc/dvips/base/8r.enc}
</usr/local/texlive/2019/texmf-dist/fonts/type1/public/amsfonts/cm/cmbx10.pfb>
""".replace(
    "\n", ""
)
